    
    #set (MOD_DIRECTORY ${CMAKE_CURRENT_BINARY_DIR}/mod)    
    
    IF (OPENMP)
         add_definitions(-DMANAGE_OMP)
    ENDIF (OPENMP)
       
   IF (WIN32)
    add_definitions(-DWINDOWS)
   ELSE (WIN32)
    add_definitions(-DLINUX)
   ENDIF (WIN32)
   
# make sure that the default is a RELEASE
	if (NOT CMAKE_BUILD_TYPE)
	  set (CMAKE_BUILD_TYPE RELEASE CACHE STRING "Choose the type of build, options are: None Debug Release." FORCE)
	endif (NOT CMAKE_BUILD_TYPE)
    
    
    # FFLAGS depend on the compiler
    # MESSAGE(FATAL_ERROR "compiler :${CMAKE_Fortran_COMPILER}")
    
    get_filename_component (Fortran_COMPILER_NAME ${CMAKE_Fortran_COMPILER} NAME)       
       
#----------------------- GFORTRAN ----------------------------------------	
	if (Fortran_COMPILER_NAME MATCHES "gfortran") #gnu
	   #OPTION OPENMP COMMUN aux compilateurs
       #--------------------------------------
       #gprof option...
       IF (GPROF)
         SET(GPROF_OPT "-pg")
       ELSE(GPROF)
         SET(GPROF_OPT "")
       ENDIF (GPROF)
   	   set (CMAKE_Fortran_FLAGS_RELEASE "${GPROF_OPT} -fopenmp -DMANAGE_OMP -O2 -DDATE_BUILD=\"'${NOW}'\"")
       set (CMAKE_Fortran_FLAGS_DEBUG   "${GPROF_OPT} -fopenmp -DMANAGE_OMP -DDATE_BUILD=\"'${NOW}'\" -fbacktrace -ggdb -fbounds-check -fcheck-array-temporaries -DDEBUG -Warray-bounds -Wsurprising -Wunderflow  -Walign-commons -fno-f2c -O0 -g")
       set (CMAKE_Fortran_FLAGS77  "")
       set (CMAKE_Fortran_FLAGS95  "")
       set (CMAKE_CXX_FLAGS_RELEASE "-O2 -fopenmp -DMANAGE_OMP")
       set (CMAKE_CXX_FLAGS_DEBUG   "-g -fopenmp -DMANAGE_OMP")
#----------------------- IFORT ----------------------------------------
	elseif (Fortran_COMPILER_NAME MATCHES "ifort") # intel
       #OPTION OPENMP COMMUN aux compilateurs
       #--------------------------------------
       SET(ADD_OPTION "${ADD_OPTION} -openmp -openmp-lib=compat -DMANAGE_OMP")
     
       FIND_PACKAGE(IntelGABayesPref) 
	   set (CMAKE_Fortran_FLAGS_RELEASE "-O4 ${ADD_OPTION} -DDATE_BUILD=\"'${NOW}'\"")
	   set (CMAKE_Fortran_FLAGS_DEBUG   "-check all -DDEBUG ${ADD_OPTION} -O0 -g -fp-stack-check -traceback -debug all -DDATE_BUILD=\"'${NOW}'\"")
	   
	   set (CMAKE_Fortran_FLAGSF77 "-f77rtl")
	   set (CMAKE_Fortran_FLAGSF95 "-fpp -stand f95 -free")
	   
	   set (CMAKE_CXX_FLAGS_RELEASE "-O2 -openmp -openmp-lib=compat -DMANAGE_OMP")
       set (CMAKE_CXX_FLAGS_DEBUG   "-g -openmp -openmp-lib=compat -DMANAGE_OMP")
	   
#----------------------- G95 ----------------------------------------
	elseif (Fortran_COMPILER_NAME MATCHES "g95") 
	    set (ADD_OPTION "-DLINUX")

        IF (MPI_FOUND)
         SET(ADD_OPTION "${ADD_OPTION} -DMPI_ACTIVE")
       ENDIF (MPI_FOUND)
       
	   set (CMAKE_Fortran_FLAGS_RELEASE "-O3  -fno-second-underscore ${ADD_OPTION}")
	   set (CMAKE_Fortran_FLAGS_DEBUG   "-O0  -fno-second-underscore -g -fbounds-check -ftrace=frame -fzero -DDEBUG ${ADD_OPTION}")
	   set (CMAKE_Fortran_FLAGSF77 "")
	   set (CMAKE_Fortran_FLAGSF95 "-cpp")
	   
	    #linker option
       set(LIBS ${LIBS} -lf95)
	   
#----------------------- F95 ----------------------------------------
    elseif (Fortran_COMPILER_NAME MATCHES "f95") # NAG
       find_library(NAG_LIBRARY NAMES nag_nag nag nag64)
       IF ( NOT NAG_LIBRARY )
              MESSAGE(FATAL_ERROR "$$$$$$$$ Can not find NAG library $$$$$$$$$ ")
       ELSE ( NOT NAG_LIBRARY )
              MESSAGE(STATUS "******************************************* ")
              MESSAGE(STATUS "************** NAG library **************** ")
              MESSAGE(STATUS "******************************************* ")
              MESSAGE(STATUS "Find NAG: ${NAG_LIBRARY}")
              #add_library(nag_nag SHARED ${NAG_LIBRARY})
              find_library(PHREAD_LIBRARY NAMES pthread)
              set(LIBS ${LIBS} ${NAG_LIBRARY} ${PHREAD_LIBRARY})
              IF (NOT PHREAD_LIBRARY)
                MESSAGE (FATAL_ERROR "Can not find library pthread")
              ENDIF (NOT PHREAD_LIBRARY)
              add_definitions(-DHAVE_LIBNAG)
       ENDIF ( NOT NAG_LIBRARY )

	   set (CMAKE_Fortran_FLAGS_RELEASE "-O4 -maxcontin=200  -gline -colour -ieee=full -DDATE_BUILD=\"'${NOW}'\"")
	   set (CMAKE_Fortran_FLAGS_DEBUG   "-nan -maxcontin=200  -DDEBUG -w=obs -g -C -gline -colour -ieee=stop")
	   
       set (CMAKE_Fortran_FLAGSF77 "-dusty -dcfuns")
	   set (CMAKE_Fortran_FLAGSF95 "-f95")

       SET(CMAKE_SHARED_LIBRARY_LINK_Fortran_FLAGS "")
	   SET(CMAKE_SKIP_BUILD_RPATH  TRUE)


    else (Fortran_COMPILER_NAME MATCHES "gfortran")
	  message ("CMAKE_Fortran_COMPILER full path: " ${CMAKE_Fortran_COMPILER})
	  message ("Fortran compiler: " ${Fortran_COMPILER_NAME})
	  message (FATAL_ERROR "The compiler no matches with this configuration file.")
#	  set (CMAKE_Fortran_FLAGS_RELEASE "-O2 -Wp,-DDATE_BUILD=\"'${NOW}'\"")
#	  set (CMAKE_Fortran_FLAGS_DEBUG   "-O0 -g -Wp,-DDEBUG,-DDATE_BUILD=\"'${NOW}'\"")	 
    endif (Fortran_COMPILER_NAME MATCHES "gfortran")
    
    set (CMAKE_Fortran_FLAGS_DEBUG "${CMAKE_Fortran_FLAGS_DEBUG} ${CMAKE_Fortran_FLAGSF95}")
    set (CMAKE_Fortran_FLAGS_RELEASE  "${CMAKE_Fortran_FLAGS_RELEASE} ${CMAKE_Fortran_FLAGSF95}")
    
    
    
       #affichage
    if (CMAKE_BUILD_TYPE STREQUAL "Debug")
	  MESSAGE(STATUS "*****************DEBUG************************** ")
	  MESSAGE(STATUS ${CMAKE_Fortran_FLAGS_DEBUG})
	else (CMAKE_BUILD_TYPE STREQUAL "Debug")
	  MESSAGE(STATUS "*****************RELEASE************************** ")
	  MESSAGE(STATUS ${CMAKE_Fortran_FLAGS_RELEASE})
	endif (CMAKE_BUILD_TYPE STREQUAL "Debug")
	
	
	
    
