program main
    use mod_struct
    use mod_lecture
    use mod_calcul
    use mod_random

#ifdef MANAGE_OMP
    use omp_lib
#endif

    implicit none

    integer       , parameter      :: outFileId = 998 ! identifiant UNIT pour ecrire dans le fichier de sortie

    integer                        :: nb,n,ios,sizeId,nubeta,icycle,jcycle,j
    integer                        :: nbCycleByChain,nbCycleBurnIn,nbChain,numcycle,i,selectGeneratorUser
    integer                        :: processusId,status
    character(len=LENGTH_MAX_FILE) :: paramfile,filey,filex,outFile
    character(len=LEN_L)           :: v
    logical                        :: ok,vprint,randompi
    type(DATASET)                  :: data
    type(Solution)                 :: estim
    character(len=LEN_L),dimension(:),allocatable :: lanim
    real(kind=dp)                  :: deuxPQBarre,varA,varE,pi,mean,var,t1,t2
    character(len=100)             :: phrase
    integer ,dimension(8)          :: values

    !pour chque process parallele on stocke les points
    type(DistributionParameter)   ,dimension(:) , allocatable :: distrib
    !reunion de tous les points generer par tous les processus
    type(DistributionParameter)                               :: distribGather

    nb = command_argument_count()
    if ( nb < 1 ) then
      write (0,*) 'The parameter file are unknown.'
      write (0,*) K_NUMANIM
      write (0,*) K_NUMMARK
      write (0,*) K_TRAIT
      write (0,*) K_GENOTYPE
      write (0,*) K_PHENO_GENOTYPE
      !write (0,*) K_UNKNOWN
      write (0,*) K_VARA
      write (0,*) K_VARE
      write (0,*) K_PI
      write (0,*) K_RANDOMPI
      write (0,*) K_NUBETA
      write (0,*) K_NUMBERCHAIN
      write (0,*) K_CHAINLENGTH
      write (0,*) K_NUMCYCLE
      write (0,*) K_BURNIN
      write (0,*) K_OUTPUT
      write (0,*) K_RANDOMGENERATOR
      stop
    end if

    call GETARG(1,paramfile)

    vprint = .false.
    if ( nb >= 2 ) then
       phrase=''
       call GETARG(2,phrase)
       print *,'add options?:',phrase
       if ( trim(phrase) == 'verbose' .or. trim(phrase) == 'v') vprint = .true.
    end if

    !lecture du fichier de parametre
    call read_param(paramfile)
    !verifier si tous les parametres utilisateurs sont presents
    call check_keys

    !Lecture du nombre d animaux a prendre en compte
    call get_key_value(K_NUMANIM,v)

    data%na = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_NUMANIM)
        stop
    end if

    !Lecture du nombre de marqueur a considerer
    call get_key_value(K_NUMMARK,v)
    data%nmk = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_NUMMARK)
        stop
    end if

    call get_key_value(K_CHAINLENGTH,v)
    nbCycleByChain = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_CHAINLENGTH)
        stop
    end if

    call get_key_value(K_BURNIN,v)
    nbCycleBurnIn = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_BURNIN)
        stop
    end if

    !Set generator random
    ! 1 => Cooked random (methode collecte et arrange pour fonctionner avec openmp)
    ! 2 => Tina's random lib : lib dedie a la generation de nombre pour les programme monte carlo parallel
    selectGeneratorUser=COOKED ! par defaut
    if ( key_exist(K_RANDOMGENERATOR) ) then
      call get_key_value(K_RANDOMGENERATOR,v)
      selectGeneratorUser = get_int(v,ok)
      if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_RANDOMGENERATOR)
        stop
      end if
    end if

#ifdef MANAGE_OMP

    call get_key_value(K_NUMBERCHAIN,v)
    nbChain = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_NUMBERCHAIN)
        stop
    end if

    if ( nbChain <= 0 .or. nbChain > omp_get_max_threads() ) then
        write (*,*) "Can not allocate the number of thread ask by the user :",trim(K_NUMBERCHAIN),'=',nbChain
        write (*,*) "Maximum :",omp_get_max_threads()
        write (*,*) "End."
        stop
    end if

    !allocation of the general structure distrib

    write (*,*) "  +++ OpenMP +++ "
    write (*,*) "================="
    write (*,*) " nbChain       = ",nbChain
    write (*,*) " max available = ",omp_get_max_threads()

    call omp_set_num_threads(nbChain)

#else
    nbChain = 1
    write (*,*) " ** Compilation without OpenMP *** "
    write (*,*) "==================================="
    write (*,*) " nbChain = 1 "

#endif

    call init_random_mode(selectGeneratorUser)
    allocate (distrib(nbChain))

    if ( nbCycleBurnIn >= nbCycleByChain ) then
      write (0,*) '** chain length too small. **'
      stop
    end if

    call get_key_value(K_NUMCYCLE,v)
    numcycle = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_NUMCYCLE)
        stop
    end if

    if ( numCycle >= ( nbCycleByChain - nbCycleBurnIn )  ) then
          write (0,*) '** numCycle too big. **'
          stop
    end if


    if ( nbCycleByChain <= 0 .or. nbCycleBurnIn <= 0 .or.  numcycle <= 0 ) then
        write (0,*) 'Bad definition of mcmc keys definition:',nbCycleByChain,nbCycleBurnIn,numcycle
        stop
    end if

    n = (nbCycleByChain - nbCycleBurnIn+1) / numcycle
    !le nombre de point total
    distribGather%nb = n * nbChain
    call new(distribGather,distribGather%nb,data%nmk,nbChain)

    do i=1,nbChain
         distrib(i)%nb = n
    end do

    !------------------------------------------------------------------------------------
    !on recupere le fichier de sortie
    if ( .not. key_exist(K_OUTPUT) ) then
       write (0,*) 'Bad definition of :',trim(K_OUTPUT)
       stop
    end if

    call get_key_value(K_OUTPUT,outFile)

    open(UNIT=outFileId,file=trim(outFile), form="formatted",recl=2**8,iostat=ios)

    if ( ios /= 0 ) then
      write (0,*) "Can not open output file:",trim(outFile)
      stop
    end if

    !----------------------------------------------------------------------------------------------------------------------
    !allocation des structure de donnee
    call new(data)

    !SI un seul fichier contenant les performance et les genotype au format du script R
    if ( key_exist(K_PHENO_GENOTYPE) ) then
      call get_key_value(K_PHENO_GENOTYPE,filex)
      call read_x_y(filex,data)
    !Si deux fichiers phenotypes et genotype
    else if ( key_exist(K_TRAIT) .and. key_exist(K_GENOTYPE) ) then

      call get_key_value(K_TRAIT,filey)
      call get_key_value(K_GENOTYPE,filex)

      !ce tableau sert de clef pour associer le phenotype au genotype de chaque individu
      allocate (lanim(data%na))

      !Lecture du fichier de performance
      print *,"reading phenotype file..."
      call read_y(filey,data,lanim)
      print *,"ok"

      !Lecture du fichier de genotype
      print *,"reading genotype file..."
      call read_x(filex,data,lanim)
      print *,"ok"

      call check_data(data,lanim)

      deallocate(lanim)
    else !pas d autre option => arret

      write (0,*) 'None information about phenotype and genotype!'
      stop
    end if

    !------------------------------------------------------------------------------------------------------------------------

    ! on recupere la variance genetique total parametre par l utilisateur
    call get_key_value(K_VARA,v)
    varA = get_real(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_VARA)
        stop
    end if

    ! on recupere la variance residuelle parametre par l utilisateur
    call get_key_value(K_VARE,v)
    varE = get_real(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_VARE)
        stop
    end if

    ! on recupere la proba initiale pi parametre par l utilisateur
    call get_key_value(K_PI,v)
    pi = get_real(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_PI)
        stop
    end if

    ! doit on fixer PI
    call get_key_value(K_RANDOMPI,v)
    randompi = get_bool(v)

    ! on recupere le ddl de la loi de CHI2 inverse de varbeta : nubeta
    call get_key_value(K_NUBETA,v)
    nubeta = get_int(v,ok)
    if ( .not. ok ) then
        write (0,*) 'Bad definition of :',trim(K_NUBETA)
        stop
    end if

    call DATE_AND_TIME(values=values)
    write (outFileId,fmt='("Start ",i2,":",i2,":",i2)') values(5),values(6),values(7)
    write (*,fmt='("Start ",i2,":",i2,":",i2)') values(5),values(6),values(7)

    !!! * Attention les commentaires debutant par OMP sont interpreter
    !!!   par le compilateur pour la parallelisation
    !!! ================================================================

    !! - Par defaut l'ensemble des variable est prive sauf ceux specifier dans le mot clef SHARED
    !! - il y a nbChian processus

    !! debut de la parallelisation

    !$OMP PARALLEL DEFAULT(PRIVATE) &
    !$OMP SHARED(nbChain,vprint,distrib,varA,varE,pi,data,nbCycleBurnIn,nbCycleByChain) &
    !$OMP SHARED(numcycle,nubeta,distribGather,outFile,randompi)
#ifdef MANAGE_OMP
    processusId = omp_get_thread_num()+1
#else
    processusId = 1
#endif

    call init_random(nbChain,(processusId-1),data)
    !initialisation
    call new(distrib(processusId),distrib(processusId)%nb,data%nmk,1)

    !allocation des effets SNP a estimer
    call new(estim,data%nmk)
    call init_bayes(data,varA,varE,pi,estim,deuxPQBarre)
  !  call print_debug(data,estim)

#ifdef MANAGE_OMP
    t1=omp_get_wtime();
#else
    t1=second();
#endif
    print *,'burnin....[',processusId,']'

    !chauffe/burn in....
    do icycle=1,nbCycleBurnIn
      call bayesCpi(data,varA,deuxPQBarre,nubeta,estim,randompi)
      if ( vprint ) then
         write( *,* ) '[',icycle,']','[Proc',processusId,'] :',&
           'vare:',estim%vare, ' vareffect:', estim%varbeta ,' pi:',estim%pi, ' nloci:',sum(estim%deltaOne)
      else if (mod(icycle,200) == 0 ) then
          write( *,* ) '[',icycle,']','[Proc',processusId,'] :',&
           'vare:',estim%vare, ' vareffect:', estim%varbeta ,' pi:',estim%pi, ' nloci:',sum(estim%deltaOne)
      end if
      !call print_debug(data,estimIn,estimOut)
    end do
#ifdef MANAGE_OMP
    var = (omp_get_wtime() - t1) / real(nbCycleBurnIn)
#else
    var = (second() - t1) / real(nbCycleBurnIn)
#endif
    jcycle=0
    distrib(processusId)%nb = 0

#ifdef MANAGE_OMP
    print *,"elapsed time = ",omp_get_wtime() - t1,'[',processusId,']'
#else
    print *,"elapsed time = ",second() - t1,'[',processusId,']'
#endif

    print *,'duration :',var*real(nbCycleByChain-nbCycleBurnIn)/60.d0,'min','[Proc',processusId,']'

    do icycle=nbCycleBurnIn,nbCycleByChain
      jcycle = jcycle +1
      call bayesCpi(data,varA,deuxPQBarre,nubeta,estim,randompi)
      !call print_debug(data,estimIn,estimOut)
      if ( jcycle == numcycle ) then
        distrib(processusId)%nb = distrib(processusId)%nb + 1
        distrib(processusId)%allVare(distrib(processusId)%nb) = estim%vare
        distrib(processusId)%allVarbeta(distrib(processusId)%nb) = estim%varbeta
        distrib(processusId)%allPi(distrib(processusId)%nb) = estim%pi
        distrib(processusId)%SumBeta(1,:) = distrib(processusId)%SumBeta(1,:) + estim%beta
        distrib(processusId)%SumSquareBeta(1,:) = distrib(processusId)%SumSquareBeta(1,:) + estim%beta*estim%beta
        distrib(processusId)%nbDeltaTirage(1,:) = distrib(processusId)%nbDeltaTirage(1,:) + estim%deltaOne
        jcycle = 0
      end if
      if ( vprint ) then
         write( *,* ) '[',icycle,']','[Proc',processusId,'] :',&
           'vare:',estim%vare, ' vareffect:', estim%varbeta ,' pi:',estim%pi, ' nloci:',sum(estim%deltaOne)
      else if (mod(icycle,200) == 0 ) then
          write( *,* ) '[',icycle,']','[Proc',processusId,'] :',&
           'vare:',estim%vare, ' vareffect:', estim%varbeta ,' pi:',estim%pi, ' nloci:',sum(estim%deltaOne)
      end if
    end do

    !on ecrit dans un fichier les points de chaque processus
    open(UNIT=11+processusId,file=trim(outFile)//"."//trim(str(processusId)), form="formatted",recl=2**8,iostat=ios)

    if ( ios == 0) then
      write (11+processusId,*) "vare varbeta pi"
      do i=1,distrib(processusId)%nb
       write (11+processusId,*)  distrib(processusId)%allVare(i),distrib(processusId)%allVarbeta(i),distrib(processusId)%allPi(i)
      end do
      close (11+processusId)
    else
       write (*,*) "Proc [",processusId,"] can not open file :",trim(outFile)//"."//trim(str(processusId))
    end if

    !Gather (on reunit tous les points)
    !----------------------------------
    distribGather%allVare((distrib(processusId)%nb*(processusId-1))+1:distrib(processusId)%nb*processusId) = &
      distrib(processusId)%allVare(1:distrib(processusId)%nb)
    distribGather%allVarbeta(distrib(processusId)%nb*(processusId-1)+1:distrib(processusId)%nb*processusId) = &
      distrib(processusId)%allVarbeta
    distribGather%allPi(distrib(processusId)%nb*(processusId-1)+1:distrib(processusId)%nb*processusId) = &
      distrib(processusId)%allPi
    distribGather%SumBeta(processusId,:) = distrib(processusId)%SumBeta(1,:)
    distribGather%SumSquareBeta(processusId,:) = distrib(processusId)%SumSquareBeta(1,:)
    distribGather%nbDeltaTirage(processusId,:) = distrib(processusId)%nbDeltaTirage(1,:)

    call release_distribution_parameter(distrib(processusId))
    call release_solution(estim)
    call release_inside_parallel_random_mode

    !$OMP END PARALLEL
    !! Fin parallelisation
    !! *******************

    write (outFileId,*) 'nbpoint      :',distribGather%nb
    call get_meanvar(distribGather%nb,distribGather%allVare,mean,var)
    write (outFileId,*) 'VARE mean    :',mean,' s.d:',sqrt(var)
    call get_meanvar(distribGather%nb,distribGather%allVarbeta,mean,var)
    write (outFileId,*) 'VARBETA mean :',mean,' s.d:',sqrt(var)
    call get_meanvar(distribGather%nb,distribGather%allPi,mean,var)
    write (outFileId,*) 'PI mean      :',mean,' s.d:',sqrt(var)

    write (outFileId,*) 'SNP    MeanBeta      S.D      Nb Delta_1    NbDelta_1 chain1 ....'
    do i=1,data%nmk
      !call get_meanvar(distribGather%nb,distribGather%allBeta(:,i),mean,var)
      mean = sum(distribGather%SumBeta(:,i))/real(distribGather%nb)
      var  = sum(distribGather%SumSquareBeta(:,i))/real(distribGather%nb) - (mean*mean)
      write (outFileId,*) i,mean,sqrt(var),sum(distribGather%nbDeltaTirage(:,i)),(distribGather%nbDeltaTirage(j,i),j=1,nbChain)
    end do

    deallocate (distrib)
    call release_distribution_parameter(distribGather)
    call release(data)

    call DATE_AND_TIME(values=values)
    write (outFileId,fmt='("End ",i2,":",i2,":",i2)') values(5),values(6),values(7)
    write (*,fmt='("End ",i2,":",i2,":",i2)') values(5),values(6),values(7)

    contains

      subroutine print_debug(data,estimIn)
        type(DATASET)   ,intent(in)    :: data   ! le jeu de donnee
        type(Solution)  ,intent(in)    :: estimIn  ! l'estimation de la solution

        print *,"----------------------------------------------"
        print *,"** estimeIn **"
        print *,"mu:",estimIn%mu
        print *,"Beta(1:10)",estimIn%beta(:min(10,data%nmk)),' sum:',sum(estimIn%beta)
        print *,"varbeta:",estimIn%varbeta
        print *,"vare:",estimIn%vare
        print *,"pi:",estimIn%pi

      end subroutine


end program main
