module mod_calcul
    use mod_struct            ! module structure de donnee
    use mod_random

    public :: get_freq_all ! donne la frequence de l allele 1 du locus j
    public :: init_bayes   ! Initilisation des parametres
    public :: bayesCpi
    public :: get_meanvar  ! calcul la moyenne et la variance de la liste de point ldata

    integer          , private :: nb_stream_over_unknown_genotype = 0
    !$omp threadprivate (nb_stream_over_unknown_genotype)


    contains

      !donne la frequence de l allele 1 du locus j
      subroutine get_freq_all(data,j,p)
         type(DATASET)   ,intent(in)    :: data ! le jeu de donnee
         integer         ,intent(in)    :: j    ! le locus j
         real(kind=dp)   ,intent(out)   :: p    ! la frequence de l allele 1 calcule

         integer :: i
         real(kind=dp) :: countUn

         countUn = 0.d0
         do i=1,data%na ! pour tous les animaux

           if ( data%x(i,j) == CODE_11 ) then
               countUn = countUn + 2.d0
           else if ( data%x(i,j) == CODE_12 ) then
               countUn = countUn + 1.d0
           end if

         end do

         p = countUn / real(2*data%nmk)

      end subroutine get_freq_all


      ! Initilisation des parametres
      ! mu  : moyenne des performances
      !beta : 0 pour tous les marqueurs
      !
      subroutine init_bayes(data,varA,varE,pi,sol,deuxPQBarre)
        type(DATASET)   ,intent(in)    :: data ! le jeu de donnee
        real(kind=dp)   ,intent(in)    :: varA ! variance parametrer par l utilisateur
                                               !(variance genetique totale explique par tous les SNP)
        real(kind=dp)   ,intent(in)    :: varE ! variance residuelle donne par l utilisateur
        real(kind=dp)   ,intent(in)    :: pi   ! Pi valeur compris entre 0 et 1
        type(Solution)  ,intent(out)   :: sol                ! initialisation de l estimation
        real(kind=dp)   ,intent(out)   :: deuxPQBarre        ! entite reutiliser par la resolution (evite de recalculer le resultat)

        integer       :: i,j
        real(kind=dp) :: s
        real(kind=dp),dimension(data%nmk) :: fre

        !calcul de la moyenne
        sol%mu=0.d0
        do i=1,data%na
            sol%mu = sol%mu + data%y(i)
        end do
        sol%mu = sol%mu / real(data%na)
        !beta
        sol%beta  = 0

        !varbeta
        !
        !   varBeta = varA / ( K (1-PI)2PQ)   2PQ=2*1/K*SUM Pj(1-Pj)  Pj frequence de allele 1 au locus j
        s = 0
        do j=1,data%nmk

          if ( data%availXGlob(j) ) then
             fre(j) = sum(data%x(:,j))/(2.d0*real(data%na)) ! j ai pris le calcul du script R....
          else
           fre(j) = 0.d0
           do i=1,data%na
            if ( data%availXAnim(i,j) ) then
             fre(j) = fre(j) + data%x(i,j)
            end if
           end do
           fre(j) = fre(j)/(2.d0*(count(data%availXAnim(:,j))))
          end if

          !call get_freq_all(data,j,p)
          !s = p*(1-p) + s
        end do

        fre = fre*(1-fre)

        deuxPQBarre =   2 * (sum( fre ) / real(data%nmk))
       !vara_snp*(nmarkers*(1-pi)*mean2pq)
        sol%varbeta = varA  / ( real(data%nmk) * (1-pi) *deuxPQBarre)
       ! s = varA  * ( data%nmk*(1-pi)*deuxPQBarre)

        sol%vare = varE
        sol%pi = pi

        nb_stream_over_unknown_genotype = data%na*data%nmk - count( data%availXAnim(:,:) )

      end subroutine init_bayes

      ! calcul du vecteur de performance corrige
      subroutine get_vect_estim(data,estim,Ycorr)
        type(DATASET)                     ,intent(in)    :: data         ! le jeu de donnee
        type(Solution)                    ,intent(in)    :: estim        ! l'estimation de la solution en entree
        real(kind=dp) ,dimension(data%na) ,intent(out)   :: Ycorr

        integer :: i
        Ycorr = 0.d0
        do i=1,data%na ! parcours les individus
          !calcul de
          !            K
          ! Y - 1Mu - SUM XjBjDj
          !           j=1
          Ycorr(i) = data%y(i) - estim%mu - sum( (data%x(i,:) * estim%beta) )
       !   print *,sum(Ycorr), Ycorr(i), estim%mu,sum( (data%x(i,:) * estim%beta) )
        end do

      end subroutine get_vect_estim


      subroutine bayesCpi(data,varA,deuxPQBarre,nubeta,estim,piRandom)
        type(DATASET)   ,intent(in)    :: data   ! le jeu de donnee
        real(kind=dp)   ,intent(in)    :: varA ! variance parametrer par l utilisateur
        real(kind=dp)   ,intent(in)    :: deuxPQBarre
        integer         ,intent(in)    :: nubeta ! nb ddl a priori de la loi de CHI2 inverse de varbeta
        type(Solution)  ,intent(inout) :: estim  ! l'estimation de la solution
        logical         ,intent(in)    :: piRandom ! Vrai => tirage de pi sinon garde estim%pi garde sa valeur d entree

        integer                            :: i,j,nbloci,id_chisq,id_norm,id_unif,id_beta
        real                               :: u
        real(kind=dp) ,dimension(data%na)  :: Ycorr
        real ,dimension(data%na)  ::Ycorr2
        real                , dimension(1) :: meanv,val
        real              , dimension(1,1) :: covm
        real(kind=dp)                      :: tp,xiyi,xixi,xiyisquare,logt0,logt1,xixivare,xixixixivarb,m,f,logPi,logPiComp,v0,v1
        real(kind=dp)                      :: lhs,invlhs,mean,sig,aa,bb

        !init des id stream
        id_chisq=1
        id_norm=1
        id_unif=1
        id_beta=1

        !STEP 1 : tirage de la variance residuelle
        !------------------------------------------
        !on tire une V.A (u) dans une loi de CHI2 à N+3 ddl
        u = random_chisq(numRandom=id_chisq,ndf=data%na+3)
        id_chisq = id_chisq + 1

        !on recupere le vecteur de performance
        call get_vect_estim(data,estim,Ycorr)

        !affectation de la variance residuelle
        estim%vare = dot_product(Ycorr,Ycorr) / u
        !print *,estim%vare,dot_product(Ycorr,Ycorr) ,u,estim%mu
       ! estim%vare =  chin(dot_product(Ycorr,Ycorr),dble(data%na+3))
        !detection d'un overflow....
        if ( estim%vare /= estim%vare ) then
           write (*,*) "stop with an overflow...."
           write (*,*) "Ycorr(1:10) [",Ycorr(1:10),']'
           write (*,*) "sum(Ycorr,Ycorr):",sum(Ycorr)
           write (*,*) "dot_product(Ycorr,Ycorr):",dot_product(Ycorr,Ycorr)
           write (*,*) "u:",u
           stop
        end if

        !STEP 2 : tirage de la moyenne Mu
        !--------------------------------
        Ycorr = Ycorr + estim%mu
        mean  = ( sum( Ycorr ) ) / real(data%na)
        sig   = estim%vare / real(data%na)
        ! correction 29/07/2010
        sig = sqrt(sig)

        estim%mu = normal(numRandom=id_norm,mean=mean,sigma=sig)
        id_norm = id_norm + 1

        Ycorr = Ycorr - estim%mu


        logPi = log(estim%pi)
        logPiComp = log(1-estim%pi)

        !STEP 3 : Pour tous les SNP i tirage des DELATi BETAi
        !-----------------------------------------------------
        nbloci = 0

        do i=1,data%nmk

          !10/08/2010 --  Gestion des donnees manquantes --
          k=0
          ! si il existe au moin une donnee manquante pour le marqueur i
          if ( .not. data%availXGlob(i)) then
             ! on infere pour tous les animaux qui on un genotype manquant une valeur sachant la proba des genotypes calculé a partir de la population
             do ind=1,data%na
              if ( .not. data%availXAnim(ind,i) ) then
                 ! nous avons calcule la proba de chaque genotype prealablement
                 ! nous allons tire un real u si celui ci :
                 !  * 0              <=  u < proba(CODE_11)                     => genotype 11
                 !  * proba(CODE_11) <=  u < proba(CODE_11) + proba(CODE_12)    => genotype 12
                 !  * proba(CODE_12) <=  u < 1                                  => genotype 22

                 !tirage
                 call rand_number(numRandom=id_unif+k,value=u)
                 k = k + 1
                 if ( u < data%probGenotypeCase(i,1) ) then
                   data%x(ind,i) = CODE_11
                 else if ( u < data%probGenotypeCase(i,1) + data%probGenotypeCase(i,2)) then
                   data%x(ind,i) = CODE_12
                 else
                   data%x(ind,i) = CODE_22
                 end if
              end if
             end do
          end if

!          stop

          !on enleve la correction de l effet concerne
          Ycorr = Ycorr + data%x(:,i) * estim%beta(i)
          xixi = dot_product(data%x(:,i),data%x(:,i))

          if ( xixi /= 0 ) then
           xiyi = dot_product(data%x(:,i),Ycorr)
           xiyisquare = xiyi * xiyi
           xixivare = xixi*estim%vare
           logt0 = -0.5*( log(xixivare)+(xiyisquare/xixivare) ) + logPi

           xixixixivarb=xixi*xixi*estim%varbeta
           logt1 = -0.5*( log(xixivare + xixixixivarb) + (xiyisquare/(xixivare+xixixixivarb)) ) + logPiComp

           !Translation R
           !--------------
           !v0 = xixi*estim%vare
           !v1 = xixi*xixi*estim%varbeta + v0

           !logt0 = -0.5*(log(v0) + (xiyi*xiyi)/v0) + logPi;
      	   !logt1 = -0.5*(log(v1) + (xiyi*xiyi)/v1) + logPiComp;
           !FIN translation R
           !------------------

       	   m = max(logt0,logt1)
      	   f = exp( logt1 - m - log( exp(logt0-m) + exp(logt1-m)) )

           !tirage
           call rand_number(numRandom=id_unif+nb_stream_over_unknown_genotype+i-1,value=u)

           if ( u < f ) then
            !on memorise l information :"nombre de tirage a 1 du delta i"
            estim%deltaOne(i) = 1
            !on tire une valeur u
            tp = estim%vare/estim%varbeta
            mean = xiyi / ( xixi + tp )
            sig = estim%vare / ( xixi + tp )
            ! correction 29/07/2010
            sig = sqrt(sig)

            !Translation R
            !--------------
            !lhs = xixi/estim%vare + 1.d0/estim%varbeta;
      	    !invLhs = 1.d0/lhs;
            !mean = invLhs*xiyi/estim%vare
            !sig  = sqrt(invLhs)
            !FIN translation R
            !-----------------

            estim%beta(i) = normal(numRandom=id_norm + i - 1,mean=mean,sigma=sig)
            !estim%beta(i) = random_normal(real(mean),real(sig))
            Ycorr = Ycorr -  data%x(:,i)*estim%beta(i)
           else
            estim%deltaOne(i)   = 0
            estim%beta(i) = 0.d0
           end if
          end if
        end do
        id_norm = id_norm + data%nmk

        nbloci = sum(estim%deltaOne)

        !STEP 4 : Tirage de la variance des effets SNP : varbeta
        !-------------------------------------------------------
        !on tire une V.A (u) dans une loi de CHI2 à N+3 ddl
        u = random_chisq(numRandom=id_chisq,ndf=nbloci+nubeta)
        id_chisq = id_chisq + 1
        tp = dot_product(estim%beta,estim%beta) + ( (nubeta - 2) * ( varA / ( real(data%nmk)*(1-estim%pi)*deuxPQBarre ))  )

        estim%varbeta = tp / u
        !estim%varbeta = chin(tp,dble(nbdelta1+nubeta))

        if ( piRandom ) then
          !STEP 5 : tirage de PI
          !tirage dans une loi beta
          aa = real(data%nmk-nbloci + 1)
          bb = real(nbloci+1)
          estim%pi = random_beta(numRandom=id_beta,aa=real(aa),bb=real(bb))
          id_beta = id_beta + 1
        end if

      end subroutine bayesCpi

      !calcul la moyenne et la variance de la liste de point ldata
      subroutine get_meanvar(n,ldata,mean,var)
        integer                     ,intent(in)    :: n           ! nb de point
        real(kind=dp)  ,dimension(n),intent(in)    :: ldata       ! les points
        real(kind=dp)               ,intent(out)   :: mean        ! moyenne
        real(kind=dp)               ,intent(out)   :: var         ! variance

        real(kind=dp) :: tp
        integer       :: i

       ! print *,ldata
        mean = sum(ldata(:n))/real(n)
        var=0
        do i=1,n
          tp = ldata(i) - mean
          var = tp*tp + var
        end do
        var = var / real(n)

      end subroutine get_meanvar


end module mod_calcul
