module mod_lecture
    use mod_struct
    implicit none

    character(len=LEN_L)     , parameter , public :: K_NUMANIM            = 'in_numanim' !number of animal
    character(len=LEN_L)     , parameter , public :: K_NUMMARK            = 'in_nummark' !number of marker
    character(len=LEN_L)     , parameter , public :: K_TRAIT              = 'in_trait'
    character(len=LEN_L)     , parameter , public :: K_GENOTYPE           = 'in_genotype'
    character(len=LEN_L)     , parameter , public :: K_PHENO_GENOTYPE     = 'in_pheno_and_genotype'
    !character(len=LEN_L)     , parameter , public :: K_UNKNOWN            = 'in_unknown_char'
    character(len=LEN_L)     , parameter , public :: K_VARA               = 'in_vara' !varGenotypic
    character(len=LEN_L)     , parameter , public :: K_VARE               = 'in_vare' !varResidual
    character(len=LEN_L)     , parameter , public :: K_PI                 = 'in_pi'
    character(len=LEN_L)     , parameter , public :: K_RANDOMPI           = 'in_random_pi'
    character(len=LEN_L)     , parameter , public :: K_NUBETA             = 'in_nubeta' !degrees freedom effect var
    character(len=LEN_L)     , parameter , public :: K_NUMBERCHAIN        = 'in_numberchain'
    character(len=LEN_L)     , parameter , public :: K_CHAINLENGTH        = 'in_chainlength' ! nombre de cycle
    character(len=LEN_L)     , parameter , public :: K_NUMCYCLE           = 'in_numcycle' ! le nombre de chaine pour prendre en compte les valeurs des estimation apres le burn in
    character(len=LEN_L)     , parameter , public :: K_BURNIN             = 'in_burnin'
    character(len=LEN_L)     , parameter , public :: K_RANDOMGENERATOR    = 'set_random_generator'
    character(len=LEN_L)     , parameter , public :: K_OUTPUT             = 'out_output'

    ! array where key and value are stored next the read
    character(len=LENGTH_MAX_FILE)   , dimension(:), private ,allocatable :: values
    character(len=LEN_L)             , dimension(:), private ,allocatable :: keys


    interface str
        module procedure iget_str, &
                         fget_str, &
                         dget_str
    end interface str

    public :: str

    !lecture des parametres
    public :: read_param
    !Verifier l existence des clefs
    public :: check_keys

    public :: get_key_value

    public :: key_exist

    public :: get_int

    public :: get_real

    public :: get_bool

    public :: read_x

    public :: read_y

    !public format du script R
    public :: read_x_y

    ! verification de base sur le jeu de donnees
    public :: check_data

    ! rempli des structures de donnees
    ! -- boolean sur la presence d'une donnee manquante pour chaque marqueur
    ! -- proba d'obtenir le genotype -1,0,1 pour chaque marqueur
    private :: fill_additional_datainfo

    contains

    !********************************************
    !  SUBROUTINE : file_exist
    !  arrete le programme si le fichier n existe
    !********************************************
    subroutine file_exist(file_in)
        character(len=*),intent(in) ::file_in
        logical::existe
        inquire( file=file_in, exist=existe)

        if(.not. existe) then
              write (0,*) 'The file ['//trim(file_in)//'] does not exist.'
              stop
      	endif
     end subroutine file_exist

    !********************************************
    !  SUBROUTINE : GET
    !  retourne la ligne courante du fichier ouvert en lecture
    !********************************************
     subroutine GET (unit,string,iostat)
           integer                , intent(in)      :: unit
           character(len=LEN_LINE) , intent(inout)  :: string
           integer , intent(out),optional           :: iostat
           integer :: eof

           eof = 0
           string=''
           read (unit,iostat=eof,fmt="(a)/") string
           if ( present(iostat) ) iostat = eof

      end subroutine GET
    !********************************************
    !  SUBROUTINE : read_param
    !  lit le fichier de parametre qui donne les options de l analyse
    !********************************************
     subroutine read_param(param)
        character(len=LENGTH_MAX_FILE),intent(in) :: param

        integer                   :: ios_id = 20
        integer                   :: n ,ios,cln,i,j,chr
        character(len=LEN_L)      :: buf
        character(len=LEN_LINE)   :: buf2

        logical :: ok,find
        real(kind=dp) :: pas_temp

        call file_exist(param)
        open(unit=ios_id,file=trim(param),action="read",form="formatted")

        n = 0
        ios = 0
        ! count the number of line to record
        do while ( ios == 0)
            read(ios_id,*,iostat=ios) buf2
            i = index (buf2,'#')
            if ( i == 1 ) cycle
            if ( ios == 0 .and. ( trim(buf2)/='') ) then
              n = n + 1
            end if
         end do

         !allocate structure
         allocate (values(n))
         allocate(keys(n))
         keys=''

         ! init structures
         rewind(ios_id)
         cln = 0! it s the current line
         n = 0
         ios=-2
         do while ( ios /= -1 )
            call GET(ios_id,buf2,iostat=ios)
            !remove comments
            i = index (buf2,'#')

            if ( i /= 0) then
              if ( i == 1 ) buf2=''
              if ( i > 1) buf2=trim(buf2(:i-1))
            end if
            !probleme case
            if (trim(buf2)=='' .or. ios /= 0 ) cycle

            cln = cln+1
            i = index (buf2,'=')! where is =
            if ( i == 0) then
                write (0,*) cln,'bad expression of "key=value"'
                stop
            end if

              ! search if the key if exist...
              do j=1,n
                 if ( keys(j) == trim(buf2(1:i-1)) ) exit
              end do

              if ( j <= n ) then
                keys(j) = trim(buf2(1:i-1))
                values(j) = trim(buf2(i+1:))
              else
                n = n+1
                keys(n) = trim(buf2(1:i-1))
                values(n) = trim(buf2(i+1:))
              end if

              print *,"KEY ["//trim(keys(n))//"] = ["//trim(values(n))//"]"
         end do

        close(ios_id)
      end subroutine read_param

    !********************************************
    !  SUBROUTINE : key_exist
    ! rend vrai si le mot clef existe
    !********************************************
      function key_exist(keysearch) result(vexist)
         character(len=*) ,intent(in)   :: keysearch

         logical                :: vexist
         integer                :: i

         if (allocated(keys)) then
          do i=1,size(keys)
           if ( trim(keys(i)) == trim(keysearch) ) then
               if ( trim(values(i)) /= '' ) then
                 vexist=.true.
                 return
               end if

           end if
          end do
         end if

         vexist=.false.
         return

      end function key_exist

      !********************************************
    !  SUBROUTINE : key_exist_stop
    ! rend vrai si le mot clef existe
    !********************************************
      subroutine key_exist_stop(keysearch)
          character(len=*) ,intent(in)   :: keysearch
          logical :: k

          k = key_exist(keysearch)

          if ( .not. k ) then
            write (0,*) 'please definied key:',trim(keysearch),' in the parameter file.'
            stop
          end if

      end subroutine key_exist_stop


    !********************************************
    !  SUBROUTINE : get_key_value
    ! retourne la valeur de l option de l utilisateur
    !********************************************
      subroutine get_key_value(keysearch,value)
         character(len=*) ,intent(in)     :: keysearch
         character(len=*) ,intent(out)    :: value

         integer                :: i

         value=''

         if (allocated(keys)) then
          do i=1,size(keys)
           if ( trim(keys(i)) == trim(keysearch)) then
             value = values(i)
             return
           end if
          end do
         end if

      end subroutine get_key_value

      !********************************************
    !  SUBROUTINE : check_keys
    ! verifie que les clef du fichier de parametre existe
    !********************************************
      subroutine check_keys
        call key_exist_stop(K_NUMANIM)
        call key_exist_stop(K_NUMMARK)
        !call key_exist_stop(K_TRAIT)
        !call key_exist_stop(K_GENOTYPE)
        !call key_exist_stop(K_UNKNOWN)
        call key_exist_stop(K_RANDOMPI)
        call key_exist_stop(K_VARA)
        call key_exist_stop(K_VARE)
        call key_exist_stop(K_PI)
        call key_exist_stop(K_NUBETA)
        call key_exist_stop(K_NUMBERCHAIN)
        call key_exist_stop(K_CHAINLENGTH)
        call key_exist_stop(K_NUMCYCLE)
        call key_exist_stop(K_BURNIN)
        call key_exist_stop(K_OUTPUT)
      end subroutine check_keys


      function get_int(string_to_convert,is_ok) result(int_value)
             character(len=*),intent(in)      :: string_to_convert
             logical,optional,intent(inout)       :: is_ok
             integer                              :: int_value
             character(len=LEN_L)                 :: word_token
             integer                              :: ios

             word_token = trim(string_to_convert)!CHAR(string_to_convert)
             read (word_token,'(I8)',IOSTAT=ios) int_value

             if ( present(is_ok) ) then
                is_ok = ( ios == 0 )
             endif

             if (  ios /= 0 ) int_value = 9999
             word_token =''
             return

       end function get_int

      function get_real(string_to_convert,is_ok) result(real_value)
             character(len=*),intent(in)          :: string_to_convert
             logical,optional,intent(inout)       :: is_ok
             real(kind=dp)                        :: real_value
             character(len=LEN_L)                 :: word_token
             integer                              :: ios

             word_token = trim(string_to_convert)!CHAR(string_to_convert)
             read (word_token,'(F8.5)',IOSTAT=ios) real_value

             if ( present(is_ok) ) then
                is_ok = ( ios == 0 )
             endif

             if (  ios /= 0 ) real_value = 9999.d0
             word_token =''
             return

       end function get_real


       function get_bool(string_to_convert) result(bool_value)
             character(len=*),intent(in)          :: string_to_convert
             logical                              :: bool_value
             character(len=LEN_L)                 :: word
             integer :: nlen,i,ic

             bool_value = .false.
             word = string_to_convert
             !get lower case
             nlen = len(word)
             do i=1,nlen
                ic = ichar(word(i:i))
                if (ic >= 65 .and. ic < 90) word(i:i) = char(ic+32)
             end do

             if ( trim(word) == 'true' ) bool_value = .true.
             if ( trim(word) == 'yes' ) bool_value = .true.
             if ( trim(word) == 'on' ) bool_value = .true.

             return

       end function get_bool

      !************************************************
      ! SUBROUTINE : iget_str
      !************************************************
          function iget_str(value_int) result(res)
             integer        ,intent(in)           :: value_int
             character(len=20)                    :: res
             integer                              :: io

             res=''
             write (res,fmt='(I8)',iostat=io) value_int
             if ( io /= 0 ) then
               res='<NO-DEF>'
             else
               res=adjustl(res)
             end if
             return

          end function iget_str

  !************************************************
      ! SUBROUTINE : fget_str
      !************************************************
          function fget_str(value) result(res)
             real           ,intent(in)           :: value
             character(len=20)                    :: res
             real (kind=dp)                       :: v

             v = value
             res = dget_str(v)

          end function fget_str


      !************************************************
      ! SUBROUTINE : dget_str
      !************************************************
          function dget_str(value) result(res)
             real (kind=dp) ,intent(in)           :: value
             character(len=20)                    :: res
             integer                              :: io

             res=''
             write (res,fmt='(f13.2)',iostat=io) value
             if ( io /= 0 ) then
               res='<NO-DEF>'
             else
               res=adjustl(res)
             end if
             return

          end function dget_str


      subroutine read_y(filey,data,lanim)
        character(len=LENGTH_MAX_FILE),intent(in)    :: filey  ! le fichier de performance
        type(DATASET)                 ,intent(inout) :: data     ! le nombre d animaux
        character(len=LEN_L),dimension(data%na),intent(out) :: lanim ! la liste des animaux

        character(len=LEN_L) :: tps
        integer :: ios_id,i,ios,n

        call file_exist(filey)
        ios_id = 99
        open(unit=ios_id,file=trim(filey),action="read",form="formatted")

        !on cherche la taille d un identifiant animal
        ios = 0
        n=1
        do while ( ios == 0 .and. (n <= data%na))
            read(ios_id,*,iostat=ios) lanim(n),data%y(n)
            !print *,n,lanim(n),data%y(n)
            if ( trim(lanim(n)) /= '' ) n = n + 1
        end do

        if ( ios > 0 )then
          close(ios_id)
          write (0,*) 'An error occurs during the treatment of the file :',trim(filey)
          stop
        end if

       close(ios_id)

      end subroutine read_y

    subroutine read_x(filex,data,lanim)
        character(len=LENGTH_MAX_FILE),intent(in)    :: filex  ! le fichier de performance
        type(DATASET)                 ,intent(inout) :: data     ! le jeu de donnee
        character(len=LEN_L),dimension(data%na),intent(in) :: lanim ! la liste des animaux

        character(len=LEN_L) :: temp
        logical  , dimension(:)  ,allocatable :: is_affect
        character(len=1) ,dimension(:,:),allocatable :: l
        logical :: continue
       ! character(len=LEN_L) ,dimension(:,:,:),allocatable :: alleles

        integer :: ios_id,ios,n,i,j


        call file_exist(filex)
        ios_id = 99
        open(unit=ios_id,file=trim(filex),action="read",form="formatted")
        ios = 0
        n=1
        allocate (l(data%nmk,2))
        allocate (is_affect(data%na))
      !  allocate (alleles(data%na,data%nmk,2))

        is_affect = .false. ! initialiser a faux, lorsque le genotype de l' animal correspondant a lanim est initialise on met a vrai dans ce tableau
        !initialisation : tous les marqueurs sont presents
        data%availXAnim = .true.

        do while ( ios == 0 .and. (count(is_affect) /= data%na) )
            read(ios_id,*,iostat=ios) temp,((l(i,j),j=1,2),i=1,data%nmk)

            if ( ios ==0 .And. (trim(temp) /= '') ) then
              !on cherche l animal correspondant
              i=0
              continue =  .true.
              do while (i<data%na .and. continue)
               i=i+1
               if ( .not. is_affect(i) ) then
                  if ( trim(lanim(i)) == trim(temp) ) then
                    continue = .false.
                  end if
               end if
              end do

              !si on a pas trouve l animal correspondant...
              if ( continue ) then
               write (0,*) 'Animal ',trim(temp),' are referenced in the genotype file but not in the phenotype file.'
              else
               is_affect(i) = .true.
               do j=1,data%nmk
                if ( l(j,1) == '1' .and. l(j,2) == '1') then
                  data%x(i,j)=CODE_11
                else if ( l(j,1) == '2' .and. l(j,2) == '2') then
                  data%x(i,j)=CODE_22
                else if ( l(j,1) /= '2' .and. l(j,1) /= '1') then
                  data%availXAnim(i,j)=.false.
                else if ( l(j,2) /= '2' .and. l(j,2) /= '1') then
                  data%availXAnim(i,j)=.false.
                else
                  data%x(i,j)=CODE_12
                end if
               end do
              end if
            end if
        end do

        call fill_additional_datainfo(data)

        if ( ios > 0 )then
          close(ios_id)
          write (0,*) 'An error occurs during the treatment of the file :',trim(filex)
          stop
        end if

       close(ios_id)
     !  deallocate (alleles)
       deallocate (is_affect)
       deallocate (l)

      end subroutine read_x

     ! Format du fichier en entree
     ! ID PERF SEXE (-1,0,1).......   -1 pour AA 0 pour aA , 1 aa

     subroutine read_x_y(file,data)
        character(len=LENGTH_MAX_FILE),intent(in)           :: file  ! le fichier de performance
        type(DATASET)                 ,intent(inout)        :: data     ! le nombre d animaux

        character(len=LEN_L)        :: v
        integer :: ios_id,i,ios,n,sizeId,vv,j
        integer ,dimension(:),allocatable :: l
        logical :: is_ok

        print *,"read file : format ID Y(ID) S M1 M2 M3 .... (MI=-1 AA |0 Aa |1 aa) "

        call file_exist(file)
        ios_id = 99
        open(unit=ios_id,file=trim(file),action="read",form="formatted")

        !on cherche la taille d un identifiant animal
        ios = 0
        allocate (l(data%nmk))
        n=1
        !initialisation : tous les marqueurs sont presents
        data%availXAnim = .true.

        do while ( ios == 0 .and. (n <= data%na))
            read(ios_id,*,iostat=ios) v,data%y(n),i,(data%x(n,j),j=1,data%nmk)
            if ( (ios == 0) .And. (trim(v)/='') ) then
             ! -1 => 0
             !  0 => 1
             !  1 => 2
             do j=1,data%nmk
              ! si autre que -1,0,1 alors donnee manquante
              if (  (data%x(n,j) == -1) .or.  (data%x(n,j) == 1) .or.  (data%x(n,j) == 0)  ) then
                data%x(n,j) = mod(data%x(n,j)+1,3)
              else
                data%availXAnim(n,j) = .false.
                print *,"unknown genotype for [",n,'] [ik:',j,']:',data%x(n,j)
                data%x(n,j)=0.d0
              end if
             end do
             n = n + 1
            end if
        end do

        call fill_additional_datainfo(data)

        if ( ios > 0 )then
          close(ios_id)
          write (0,*) 'An error occurs during the treatment of the file :',trim(file)
          stop
        end if

        print *,"End read..."

       close(ios_id)
       deallocate (l)

      end subroutine read_x_y


    subroutine check_data(data,lanim)
        type(DATASET)                          ,intent(in) :: data     ! le jeu de donnee
        character(len=LEN_L),dimension(data%na),intent(in) :: lanim ! la liste des animaux

        integer :: i
        logical :: ko = .false.

        do i=1,data%nmk
         !on verifie la sum des valeurs (/=0) genotypique des individus pour un marqueur i
         if ( sum(data%x(:,i)) == 0 ) then
             ko=.true.
             write (0,*) ' *> marker ',i,' is not informativ.'
         end if
        end do

        if ( ko ) then
        !  write(0,*) " *** please check your data *** "
        !  stop
        end if

    end subroutine check_data


    !********************************************
    !  SUBROUTINE : fill_additional_datainfo
    ! rempli des structures de donnees
    ! -- boolean sur la presence d'une donnee manquante pour chaque marqueur
    ! -- proba d'obtenir le genotype -1,0,1 pour chaque marqueur
    !********************************************
    subroutine fill_additional_datainfo(data)
      type(DATASET)                 ,intent(inout) :: data     ! le jeu de donnee
      integer :: i,ind,nb
      ! on intialise par marqueur un tableau qui permet de savoir si au moin un individu a une donnee manquante
      do i=1,data%nmk
         data%availXGlob(i)= count(data%availXAnim(:,i)) == data%na
         if ( .not. data%availXGlob(i) ) then
           nb = count(data%availXAnim(:,i))
           print *,'>marq',i,' nb unknown genotype:',data%na - nb
           data%probGenotypeCase(i,:)=0.d0

           if ( nb > 0 ) then
             do ind=1,data%na
               if ( data%availXAnim(ind,i) ) then
                 if ( data%x(ind,i) == CODE_11 ) data%probGenotypeCase(i,1) = data%probGenotypeCase(i,1) + 1.d0
                 if ( data%x(ind,i) == CODE_12 ) data%probGenotypeCase(i,2) = data%probGenotypeCase(i,2) + 1.d0
                 if ( data%x(ind,i) == CODE_22 ) data%probGenotypeCase(i,3) = data%probGenotypeCase(i,3) + 1.d0
               end if
             end do
             data%probGenotypeCase(i,1) = data%probGenotypeCase(i,1) / real(nb)
             data%probGenotypeCase(i,2) = data%probGenotypeCase(i,2) / real(nb)
             data%probGenotypeCase(i,3) = data%probGenotypeCase(i,3) / real(nb)
             print *,"Genotype porbabilities 11:",data%probGenotypeCase(i,1),' 12:',data%probGenotypeCase(i,2),&
             ' 22:',data%probGenotypeCase(i,3)
           end if

         end if

      end do

      if ( count(data%availXGlob(:)) ==  data%nmk ) &
         print *,"** All genotype are filled **"

    end subroutine fill_additional_datainfo


    !********************************************
    !  SUBROUTINE : release_lecture
    ! libere les structure de donnees de ce module
    !********************************************
      subroutine release_lecture

         deallocate (values)
         deallocate(keys)

      end subroutine release_lecture


end module mod_lecture
