module mod_random
    use mod_struct
    use mod_cooked_random , only : init_cooked_random=>init_random, &
                                   cooked_normal=>normal,&
                                   random_cooked_beta=>random_beta,&
                                   random_cooked_chisq=>random_chisq,&
                                   rand_cooked_number=>rand_number

    implicit none
    !! Precision (kind) of real defined :
    !! 15 digits and exponent range greater at least 307.
 !   integer, private , parameter :: dp    = SELECTED_REAL_KIND( 15, 307 )

    !utilisation de routines collectes et amenager pour la parallelisation OpenMP
    !ces routines sont a base d appel de la routine intrasec random_number
    !celle-ci est thread-safe (utilisable en parallele) mais pas adapte si le programme est fortement lie
    !a la randomisation (ce qui est le cas d un monte carlo)
    integer  , parameter            ,public       :: COOKED = 1

    !Tina’s Random Number Generator Library
    !Librairie specialise dans la generation de nombre pour les programme a base de monte carlo
    integer  , parameter            ,public       :: TRNG_LEAPFROG   = 2

    !Internal management
    integer  , parameter            ,public       :: START_RANDOM_GENERATOR = COOKED
    integer  , parameter            ,public       :: END_RANDOM_GENERATOR   = TRNG_LEAPFROG

    character(len=200) , parameter  ,private      :: error_dev_generator_selec = &
                               "Dev error * Random generator is not selected * (use init_random_mode function) : "

    ! Interface selectionner pour la generation des nombre aleatoire
    integer       , save            ,private      :: select_mode     = -1


    public :: init_random_mode
    public :: release_inside_parallel_random_mode
    public :: init_random
    public :: normal
    public :: random_beta
    public :: random_chisq
    public :: rand_number

    ! ** TRNG Implementation **
    ! Interface with C++ develppement of mod_trng.cc
    interface
!            subroutine init_random_mode_leap ()
!            end subroutine init_random_mode_leap

!            subroutine release_outside_parallel_random_mode_trng
!            end subroutine release_outside_parallel_random_mode_trng

            subroutine release_inside_parallel_random_mode_trng_
            end subroutine release_inside_parallel_random_mode_trng_

            subroutine init_random_leap(nbProc,whichProc,nb_chisq,nb_norm,nb_unif,nb_beta)
                integer ,intent(in) :: nbProc ! number of processor
                integer ,intent(in) :: whichProc  ! numero du processus OpenMP
                integer ,intent(in) :: nb_chisq,nb_norm,nb_unif,nb_beta  ! nombre de generateur
            end subroutine init_random_leap

            subroutine set_block_stream(numRand,blockSize)
                integer ,intent(in) :: numRand  ! nombre de generateur
                integer ,intent(in) :: blockSize ! taille du block
            end subroutine set_block_stream

            subroutine trng_normal_leap(numRand,mean,sig,res)
                 integer ,intent(in) :: numRand    ! index du random
                 real    ,intent(in) :: mean
                 real    ,intent(in) :: sig
                 real    ,intent(out) :: res
            end subroutine trng_normal_leap

            subroutine trng_random_beta_leap(numRand,aa,bb,res)
                 integer ,intent(in) :: numRand    ! index du random
                 real    ,intent(in) :: aa
                 real    ,intent(in) :: bb
                 real    ,intent(out) :: res
            end subroutine trng_random_beta_leap

            subroutine trng_random_chisq_leap(numRand,ndf,res)
                 integer ,intent(in) :: numRand    ! index du random
                 integer ,intent(in) :: ndf
                 real   ,intent(out) :: res
            end subroutine trng_random_chisq_leap

            subroutine trng_rand_number_leap(numRand,res)
                 integer ,intent(in) :: numRand    ! index du random
                 real    ,intent(out) :: res
            end subroutine trng_rand_number_leap
    end interface

    contains

     subroutine init_random_mode(select_m)
       integer ,intent(in)          :: select_m

       select_mode = select_m

     end subroutine init_random_mode

     subroutine release_inside_parallel_random_mode

        select case( select_mode )
          case(COOKED)
          case(TRNG_LEAPFROG)
            call release_inside_parallel_random_mode_trng
          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end subroutine release_inside_parallel_random_mode


     ! ---------------------------------------
     subroutine init_random(nbProc,whichProcess,data)
        integer ,intent(in)          :: nbProc,whichProcess
        type(DATASET)                 ,intent(in) :: data
        integer :: i,nb_stream_chisq,nb_stream_norm,nb_stream_unif,nb_stream_beta

        select case( select_mode )
          case(COOKED)
            call init_cooked_random
          case(TRNG_LEAPFROG)
            ! CHISQ => vare + varbeta
            nb_stream_chisq   = 2
            ! NORMAL => mu + nmk
            nb_stream_norm    = 1 + data%nmk
            ! UNIF => nb marqueur avec donne manquante * max(animal) + beta i
            nb_stream_unif =  data%na*data%nmk - count( data%availXAnim(:,:) ) + data%nmk
            ! BETA => PI
            nb_stream_beta = 1
            print *,"**> Number of stream :",nb_stream_chisq+nb_stream_norm+nb_stream_unif+nb_stream_beta
            call init_random_leap(nbProc,whichProcess,nb_stream_chisq,nb_stream_norm,nb_stream_unif,nb_stream_beta)

          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end subroutine init_random

     ! ---------------------------------------
     function normal(numRandom,mean,sigma) !returns a normal distribution
        integer        ,intent(in)  :: numRandom
        real(kind=dp)  , intent(in) :: mean,sigma
        real(kind=dp) :: normal
        real :: m,s,n
        integer :: numRan
        numRan = numRandom - 1

        select case( select_mode )
          case(COOKED)
            normal = cooked_normal(mean,sigma)
          case(TRNG_LEAPFROG)
            m = mean
            s = sigma
            call trng_normal_leap(numRan,m,s,n)
            normal = n
          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end function normal


     ! ---------------------------------------
     function random_beta(numRandom,aa, bb)
        integer        ,intent(in)  :: numRandom
        real        , intent(in)    :: aa, bb
        real :: random_beta,r
        integer :: numRan
        numRan = numRandom - 1

        select case( select_mode )
          case(COOKED)
            random_beta = random_cooked_beta(aa, bb)
          case(TRNG_LEAPFROG)
            call trng_random_beta_leap(numRan,aa,bb,r)
            random_beta = r
          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end function random_beta


     ! ---------------------------------------
     function random_chisq(numRandom,ndf)
       integer        ,intent(in)  :: numRandom
       integer, intent(in) :: ndf
       real    :: random_chisq
       integer :: numRan,ndf2
       numRan = numRandom - 1
       ndf2=ndf

       select case( select_mode )
          case(COOKED)
            random_chisq = random_cooked_chisq(ndf)
          case(TRNG_LEAPFROG)
            call trng_random_chisq_leap(numRan,ndf2,random_chisq)
          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end function random_chisq

     ! ---------------------------------------
     subroutine rand_number(numRandom,value)
        integer        ,intent(in)  :: numRandom
        real , intent(out) :: value
        integer :: numRan
        numRan = numRandom - 1

        select case( select_mode )
          case(COOKED)
            call rand_cooked_number(value)
          case(TRNG_LEAPFROG)
            call trng_rand_number_leap(numRan,value)
          case default
               write (0,*) error_dev_generator_selec,select_mode
               stop
        end select

     end subroutine rand_number



end module mod_random
