module mod_struct
   implicit none

   !! Precision (kind) of real defined :
   !! 15 digits and exponent range greater at least 307.
   integer, public, parameter :: dp    = SELECTED_REAL_KIND( 15, 307 )

   !!   Maximum length of a name file
   integer, public, parameter             :: LENGTH_MAX_FILE    = 1024
   !!   Length size of a long word
   integer, public, parameter             :: LEN_L              = 80
   !!   Maximum size length of a line readed
   integer, public, parameter             :: LEN_LINE           = 35000

   !Note kind=1 => codage sur 1 octet (8 valeurs suffise a coder 11 12 22)

   !codage de l'homozygote 1-1
   real(kind=dp)                    , parameter , public :: CODE_11 = 0.d0
   !codage de l'heterozygote 1-2
   real(kind=dp)                    , parameter , public :: CODE_12 = 1.d0
   !codage de l'homozygote 2-2
   real(kind=dp)                    , parameter , public :: CODE_22 = 2.d0

   type DATASET
     integer                                       :: na    ! nombre d animaux
     integer                                       :: nmk   ! nombre de marqueur
     !size : na
     real(kind=dp)     ,dimension(:) ,pointer       :: y     ! perfomance
     !size na,nmk
     integer           ,dimension(:,:) , pointer    :: x     ! genotype dimension a nmk le nombre de marqueur
     !size na,nmk
     logical          ,dimension(:,:) , pointer     :: availXAnim ! gestion des donnees manquante pour un animal        : true => le genotype au marqueur est valide pour l animal
     !size nmk
     logical          ,dimension(:)   , pointer     :: availXGlob ! gestion des donnees manquante pour tous les animaux : true => le genotype au marqueur est valide pour l'ensemble de la population
     ! size nmk,3
     real             ,dimension(:,:) ,pointer      :: probGenotypeCase ! proba des 3 genotypes possibles -1, 0, 1 pour la gestion des donnees manquantes. on tire selon la proba d obtenir un des 3 genotypes

   end type DATASET

   type Solution
         real(kind=dp)                            :: mu      ! moyenne generale
         real(kind=dp) ,dimension(:)   ,pointer   :: beta    ! coefficient de regression (taille : nmk)
         integer       ,dimension(:)   ,pointer   :: deltaOne ! =1 si le delta a ete tire à 1 sinon = 0
         real(kind=dp)                            :: vare    ! variance residuelle
         real(kind=dp)                            :: varbeta ! variance des beta
         real(kind=dp)                            :: pi      ! pi
   end type Solution


   type DistributionParameter
         ! ensemble des points stocker tous les N2 cycle apres le burn in
         integer                                   :: nb
         real(kind=dp) ,dimension(:)  ,pointer     :: allVare
         real(kind=dp) ,dimension(:)  ,pointer     :: allVarbeta
         real(kind=dp) ,dimension(:)  ,pointer     :: allPi
         !attention la premiere dimension correspond au nombre de processus parallele execute
         real(kind=dp) ,dimension(:,:),pointer     :: SumBeta
         real(kind=dp) ,dimension(:,:),pointer     :: SumSquareBeta
         real(kind=dp) ,dimension(:,:),pointer     :: nbDeltaTirage
   end type DistributionParameter

   !------------------------------------------------------------------------------------------------
   !  Definition des fonctions
   !
   interface new
        module procedure new_dataset, &
                         new_solution, &
                         new_distribution_parameter
   end interface new

   interface release
        module procedure release_dataset, &
                         release_solution, &
                         release_distribution_parameter
   end interface release

   !------------------------------------------
   ! Initialisation des structures de donnees
   !

   !methode generique
   public :: new
   !methodes specifiques
   public :: new_dataset
   public :: new_solution
   public :: new_distribution_parameter


   !------------------------------------------
   ! Deallocation memoire des structures de donnees

   !methode generique
   public :: release

   !methodes specifiques
   public :: release_dataset
   public :: release_solution
   public :: release_distribution_parameter


   contains

     !----------
     subroutine new_dataset(data)
       type(DATASET) , intent(inout) :: data
       integer :: ios
       character(len=20) :: fonct = "** new_dataset **"

       if ( data%na <= 0 ) then
         write (0,*) fonct,'Can not allocate dataset : data%na is not initialized.'
         stop
       end if

       if ( data%nmk <= 0 ) then
         write (0,*) fonct,'Can not allocate dataset : data%nmk is not initialized.'
         stop
       end if

       allocate ( data%y(data%na),stat=ios )

       if ( ios /= 0 ) then
         write (0,*) fonct,'Can not allocate Y data [data%na=',data%na,']'
       stop
       end if

       allocate ( data%x(data%na,data%nmk),stat=ios)

       if ( ios /= 0 ) then
         write (0,*) fonct,'Can not allocate Genotype data [data%na=',data%na,' data%nmk=',data%nmk,']'
       stop
       end if

       allocate ( data%availXAnim(data%na,data%nmk),stat=ios)

       if ( ios /= 0 ) then
         write (0,*) fonct,'Can not allocate aray availXAnim [data%na=',data%na,' data%nmk=',data%nmk,']'
       stop
       end if

       allocate ( data%availXGlob(data%nmk),stat=ios)

       if ( ios /= 0 ) then
         write (0,*) fonct,'Can not allocate aray availXGlob [data%nmk=',data%nmk,']'
       stop
       end if

       allocate ( data%probGenotypeCase(data%nmk,3),stat=ios)

       if ( ios /= 0 ) then
         write (0,*) fonct,'Can not allocate aray probGenotypeCase [data%nmk=',data%nmk,']'
       stop
       end if

       data%availXAnim = .false.
       data%availXGlob = .false.
       data%probGenotypeCase = 0.d0


     end subroutine new_dataset

     !----------
     subroutine new_solution(sol,nmk)
       type(Solution) , intent(inout) :: sol
       integer        , intent(in)    :: nmk

       integer :: status
       character(len=25) :: fonct = "** new_solution **"

       if ( nmk <= 0 ) then
          write (0,*) fonct,"NMK : the number of marker is not valid :",nmk
          stop
       end if

       allocate (sol%beta(nmk),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate beta array NMK=",nmk
         stop
       end if

       allocate (sol%deltaOne(nmk),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate deltaOne array NMK=",nmk
         stop
       end if


       sol%beta = 0.d0
       sol%deltaOne = 0

     end subroutine new_solution

     !----------
     subroutine new_distribution_parameter(distrib,nb,nmk,nbChain)
       type(DistributionParameter) , intent(inout) :: distrib
       integer                     ,intent(in)     :: nb,nmk,nbChain

       integer :: status
       character(len=45) :: fonct = "** new_distribution_parameter **"

       if ( nb <= 0 ) then
          write (0,*) fonct," NB : the number of the sample is not valid :",nb
          stop
       end if

       if ( nmk <= 0 ) then
          write (0,*) fonct,"NMK : the number of marker is not valid :",nmk
          stop
       end if

       if ( nbChain <= 0 ) then
          write (0,*) fonct,"NBCHAIN : the number of chain is not valid :",nbChain
          stop
       end if

       ! nb : taille du nombre de valeurs a recolte pour obtenir la distribution
       allocate (distrib%allVare(nb),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate allVare array NB=",nb
         stop
       end if

       allocate (distrib%allVarbeta(nb),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate allVarbeta array NB=",nb
         stop
       end if

       allocate (distrib%allPi(nb),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate allPi array NB=",nb
         stop
       end if
       allocate (distrib%SumBeta(nbChain,nmk),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate SumBeta array NB=",nmk
         stop
       end if
       allocate (distrib%SumSquareBeta(nbChain,nmk),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate SumSquareBeta array NB=",nmk
         stop
       end if
       allocate (distrib%nbDeltaTirage(nbChain,nmk),stat=status)
       if ( status /= 0 ) then
         write (0,*) fonct,"Can not allocate nbDeltaTirage array NB=",nb
         stop
       end if

       distrib%nbDeltaTirage = 0.d0
       distrib%allVare    = 0.d0
       distrib%allVarbeta = 0.d0
       distrib%allPi      = 0.d0
       distrib%SumBeta    = 0.d0
       distrib%SumSquareBeta=0.d0

     end subroutine new_distribution_parameter


     !----------
     subroutine release_dataset(data)
       type(DATASET) , intent(inout) :: data

       deallocate (data%y)
       deallocate (data%x)
       deallocate (data%availXAnim)
       deallocate (data%availXGlob)
       deallocate (data%probGenotypeCase)

     end subroutine release_dataset

     !----------
     subroutine release_solution(sol)
       type(Solution) , intent(inout) :: sol

       deallocate (sol%beta)
       deallocate (sol%deltaOne)

     end subroutine release_solution

     !----------
     subroutine release_distribution_parameter(distrib)
       type(DistributionParameter) , intent(inout) :: distrib

       deallocate (distrib%allVare)
       deallocate (distrib%allVarbeta)
       deallocate (distrib%allPi)
       deallocate (distrib%SumBeta)
       deallocate (distrib%SumSquareBeta)
       deallocate (distrib%nbDeltaTirage)

     end subroutine release_distribution_parameter


end module mod_struct
