/*
 * mod_trng.cpp
 *
 *  Created on: 11 août 2010
 *      Author: ofilangi
 */
#include <iostream>
#include <assert.h>


#include <math.h>
#include <float.h>

//    ** Random number engine **
//
//    plusieurs generateurs sont proposes
//
//---------------------------------------

//-----------------------------------------------------------------------------
// LCG : Linear congruential generator (quick and dirty generator of TRNG but FAST)
//-----------------------------------------------------------------------------
// Sequence   ::   Ri+1 = ( Ri A.Ri + b ) mod 2^64
//
// periode 2^64 si b est impair et modulo(a,4) = 1
//
//#include <trng/lcg64.hpp>
//#define GENRATOR_TYPE                  trng::lcg64

// with additionnal bit-shift transformation :
//#include <trng/lcg64_shift.hpp>
//#define GENRATOR_TYPE                  trng::lcg64_shift
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// MRGN : Multiple recurence generator based on a linear feedback shift register sequence over F(2^31-1) of depth N
//-----------------------------------------------------------------------------
// Sequence   ::  Ri = (a1 . Ri-1 + a2 . Ri-2 + ... + an . Ri-n-2 ) mod M
//

// PERIODE : 2^62
//#include <trng/mrg2.hpp>
//#define GENRATOR_TYPE                  trng::mrg2

// PERIODE : 2^93
//#include <trng/mrg3.hpp>
//#define GENRATOR_TYPE                  trng::mrg3

// PERIODE : 2^124
//#include <trng/mrg4.hpp>
//#define GENRATOR_TYPE                  trng::mrg4

// PERIODE : 2^155
//#include <trng/mrg5.hpp>
//#define GENRATOR_TYPE                  trng::mrg5

//-----------------------------------------------------------------------------
// MRGNS : Multiple recurence generator based on a linear feedback shift register sequence over F(M) of depth N, with M being a Sophie Germain Prime
//-----------------------------------------------------------------------------

// PERIODE : 2^93
//#include <trng/mrg3s.hpp>
//#define GENRATOR_TYPE                  trng::mrg3s

// PERIODE : 2^155
//#include <trng/mrg5s.hpp>
//#define GENRATOR_TYPE                  trng::mrg5s

//-----------------------------------------------------------------------------
// YARN : Yet Another Random Number sequence based on a linear feedback shift register sequence over F(2^31-1) of depth N
// based on a multiple recursive generator
//-----------------------------------------------------------------------------
// Sequence   ::  Ri = (a1 . Ri-1 + a2 . Ri-2 + ... + an . Ri-n ) mod M
//

// PERIODE : 2^62
//#include <trng/yarn2.hpp>
//#define GENRATOR_TYPE                  trng::yarn2

// PERIODE : 2^93
//include <trng/yarn3.hpp>
//define GENRATOR_TYPE                  trng::yarn3

// PERIODE : 2^124
#include <trng/yarn4.hpp>
#define GENRATOR_TYPE                  trng::yarn4

// PERIODE : 2^155
//#include <trng/yarn5.hpp>
//#define GENRATOR_TYPE                  trng::yarn5

//-----------------------------------------------------------------------------
// YARNS : Yet Another Random Number sequence based on a linear feedback shift register sequence over F(M) of depth N, with M being a Sophie Germain Prime
//-----------------------------------------------------------------------------
// PERIODE : 2^93
//include <trng/yarn3s.hpp>
//define GENRATOR_TYPE                  trng::yarn3s

// PERIODE : 2^155
//#include <trng/yarn5s.hpp>
//#define GENRATOR_TYPE                  trng::yarn5s


//Random number distribution used
//-------------------------------
#include <trng/uniform01_dist.hpp>
#include <trng/normal_dist.hpp>
#include <trng/chi_square_dist.hpp>
#include <trng/gamma_dist.hpp>


#define DEGREE_OF_FREEDOM_CHISQ_LIMIT  200

//On garde l ensemble des generateurs
static GENRATOR_TYPE          *streamsChisq  = NULL;
static GENRATOR_TYPE          *streamsNormal = NULL;
static GENRATOR_TYPE          *streamsUnif   = NULL;
static GENRATOR_TYPE          *streamsBeta   = NULL;

//Directive openMP : les tableaux streamsXXX sont locaux a chaque processus OpenMP
#pragma omp threadprivate (streamsChisq,streamsNormal,streamsUnif,streamsBeta)


extern "C" void release_inside_parallel_random_mode_trng_() {
	if ( streamsChisq ) {
	    delete[] streamsChisq ;
	    streamsChisq = NULL;
	}
	if ( streamsNormal ) {
	    delete[] streamsNormal ;
	    streamsNormal = NULL;
	}

	if ( streamsUnif ) {
        delete[] streamsUnif ;
        streamsUnif = NULL;
	 }

	if ( streamsBeta ) {
        delete[] streamsBeta ;
        streamsBeta = NULL;
     }
}

extern "C" void init_random_leap_(int*nbProc,int *whichProc,int*nbChisq,int*nbNorm,int*nbUnif,int*nbBeta) {

	assert (whichProc!=NULL);

	// LEAPFROG Method .
    // allocation of a random number engine for each random we needs

	if ( *nbChisq > 0 ) streamsChisq  = new GENRATOR_TYPE[*nbChisq];
	if ( *nbNorm > 0 ) streamsNormal = new GENRATOR_TYPE[*nbNorm];
	if ( *nbUnif > 0 ) streamsUnif   = new GENRATOR_TYPE[*nbUnif];
	if ( *nbBeta > 0 ) streamsBeta   = new GENRATOR_TYPE[2*(*nbBeta)]; // we used 2 gamma...

	int total = *nbChisq + *nbNorm + *nbUnif + 2*(*nbBeta) ;

	for (int j=0;j<*nbChisq;j++) {
		streamsChisq[j].split(total,j);
		streamsChisq[j].split(*nbProc,*whichProc);
    }
	for (int j=0;j<*nbNorm;j++) {
		streamsNormal[j].split(total,*nbChisq + j);
		streamsNormal[j].split(*nbProc,*whichProc);
    }
	for (int j=0;j<*nbUnif;j++) {
		streamsUnif[j].split(total,*nbChisq + *nbNorm + j);
		streamsUnif[j].split(*nbProc,*whichProc);
	}
	for (int j=0;j<2*(*nbBeta);j++) {
		streamsBeta[j].split(total,*nbChisq + *nbNorm + *nbUnif + j);
		streamsBeta[j].split(*nbProc,*whichProc);
	}
}


extern "C" void trng_normal_leap_(int *numRand,float*mean,float*sig,float*res) {
	trng::normal_dist<float> d(*mean, *sig);

	assert (numRand!=NULL);
	assert ((*numRand)>=0);

	*res = d(streamsNormal[*numRand]);
	//std::cout << "normal:"<<*res<<std::endl;
}



//extern "C" void trng_random_beta_leap2_(int *numRand,float*aa,float*bb,float*res) {
//
//	assert (numRand!=NULL);
//	assert ((*numRand)>=0);
//
//	// Beta (aa,bb) = Gam (1,aa) /  Gam (1,aa) + Gam (1,bb)
//	trng::gamma_dist<> gA(1.0, *aa), gB(1.0, *bb);
//	float a = gA(streamsBeta[((*numRand)*2)]);
//	float b = gB(streamsBeta[((*numRand)*2)+1]);
//	*res= a / (a+b) ;
//}

extern "C" void trng_random_chisq_leap_(int *numRand,int*ndf,float*res) {
	assert (numRand!=NULL);
    assert ((*numRand)>=0);

	if ( *ndf < DEGREE_OF_FREEDOM_CHISQ_LIMIT ) {
	  trng::chi_square_dist<> d(*ndf);
	  *res = (float)d(streamsChisq[*numRand]);
	} else {
	  //As the degrees of freedom increase, the chi-square curve approaches a normal distribution.
	  trng::normal_dist<float> d(*ndf,sqrt(2*(*ndf)));
	  *res = d(streamsChisq[*numRand]);
	}
	//std::cout << "ndf:"<< *ndf << " Num:" << *numRand << " c++ chisq:"<<*res<<std::endl;
}


extern "C" void trng_rand_number_leap_(int *numRand,float*res) {
	assert (numRand!=NULL);
    assert ((*numRand)>=0);
    //std::cout << "trng_rand_number_leap_("<<*numRand<<")"<<std::endl;
	trng::uniform01_dist<float> u;
	*res = (float)u(streamsUnif[*numRand]);
	//std::cout << "rand:"<<*res<<std::endl;
}

//C     Returns a single random deviate from the beta distribution with
//C     parameters A and B.  The density of the beta is
//C               x^(a-1) * (1-x)^(b-1) / B(a,b) for 0 < x < 1
//C
//C
//C                              Arguments
//C
//C
//C     A --> First parameter of the beta distribution
//C                         REAL A
//C
//C     B --> Second parameter of the beta distribution
//C                         REAL B
//C
//C
//C                              Method
//C
//C
//C     R. C. H. Cheng
//C     Generating Beta Variatew with Nonintegral Shape Parameters
//C     Communications of the ACM, 21:317-322  (1978)
//C     (Algorithms BB and BC)
//
//     http://www.cs.uiowa.edu/ftp/kcowles/s248/randomf/src/genbet.f

extern "C" void trng_random_beta_leap_randlib_(int *numRand,float*aa,float*bb,float*res) {
//     .. Parameters ..
//     Close to the largest number that can be exponentiated
//      REAL expmax
      float expmax=89.0;
//     Close to the largest representable single precision number
      float infnty=1.0E38;

      if ( *aa <= 0.0 || ( *bb <= 0.0 ) ) {
    	  std::cerr << "trng_random_beta_leap : bad arguments (aa:"<< *aa<<",bb:"<<*bb<<")" << std::endl ;
    	  return;
      }

      float a, b ;
      if ( *aa < *bb ) {
         a = *aa; b = *bb;
      } else {
         b = *aa; a = *bb;
      }

      //algo BB
      if ( a > 1.0 ) {
//     Alborithm BB
//
//     Initialize
//
      trng::uniform01_dist<float> uf;
      float alpha, beta, gamma,u1,u2,w,z,r,s ;
      alpha = a + b;
      beta = sqrt((alpha-2.0)/ (2.0*a*b-alpha));
      gamma = a + 1.0/beta;
      bool go40 = false;
      //do while go40
      do {
    	float v,t;
        u1 = uf(streamsBeta[((*numRand)*2)]);
//
//     Step 1
//
        u2 = uf(streamsBeta[((*numRand)*2)+1]);
        v = beta*log(u1/ (1.0-u1));
        if ( ( v > expmax) ) {
            w = infnty ;
        } else {
            w = a*exp(v);
        }
        z = pow(u1,2*u2);
        r = gamma*v - 1.3862944;
        s = a + r - w ;
//
//     Step 2
//
        bool go70 = ((s+2.609438) >= (5.0*z) ) ; //si vrai , on sort de la boucle pour calculer le resultat
        go40 = true ;
        if ( ! go70 ) {
    	  //
    	  //     Step 3
    	  //
          t = log(z);
          go70 =  s > t ;//si vrai , on sort de la boucle pour calculer le resultat
          if ( ! go70 ) {
        	  //
        	  //     Step 4
        	  //
        	  go40 = ((r+alpha*log(alpha/ (b+w))) < t) ; //on recalcule si vrai
          }
        }

      } while ( go40 ) ;

//
//     Step 5
//
      //go70
      if ( (*aa == a)) {
            *res = w/ (b+w) ;
      } else {
    	  *res = b/ (b+w) ;
      }
        return;


      } else {
//
//      ALGO BC
//
       float alpha,beta,delta,k1,k2,u1,u2,w;
       trng::uniform01_dist<float> uf;
       alpha = a + b ;
       beta = 1.0/b ;
       delta = 1.0 + a - b ;
       k1 = delta* (0.0138889+0.0416667*b)/ (a*beta-0.777778) ;
       k2 = 0.25 + (0.5+0.25/delta)*b ;
       bool go170;
       //Label 120
       do {
       u1 = uf(streamsBeta[((*numRand)*2)]);
      //
      //    Step 1
      //
       u2 = uf(streamsBeta[((*numRand)*2)+1]);
       float y,z,v;
       go170 = false;

       if (u1<0.5) {
      //
      //    Step 2
      //
          y = u1*u2 ;
          z = u1*y ;
          if ((0.25*u2+z-y)>=k1) continue ;//go120
          go170=true;
       } else { // On sort de la boucle !!
      //
      //     Step 3
      //
         z = pow(u1,2*u2);
         if ( z <= 0.25 ) {
            v = beta*log(u1/ (1.0-u1));
            if ( (v > expmax)) {
              w = infnty ;
            } else {
              w = a*exp(v);
            }
            break; //on sort go200...
         } else {
            if (z >= k2) continue ;//goto120
         }
       }
       //
       //     Step 4
       //
       //
       //     Step 5
       //
        if ( go170 ) {
         //LAbel 170
         v = beta*log(u1/ (1.0-u1)) ;
         if ( (v > expmax)) {
           w = infnty ;
         } else {
           w = a*exp(v) ;
         }
         if ((alpha* (log(alpha/ (b+w))+v)-1.3862944) < log(z)) continue ; //go120
        }
       } while ( true ) ;

       //
       //     Step 6
       //
       //Label 200
       if ( (a == *aa)) {
            *res = w/ (b+w) ;
       } else {
            *res = b/ (b+w) ;
       }

      }
}

extern "C" void trng_random_beta_leap_oldimp_(int *numRand,float*aa,float*bb,float*res) {

	    if ((*aa <= 1.0) && (*bb <= 1.0))
	    {
	        float U, V, X, Y;
	        /* Use Jonk's algorithm */
	        trng::uniform01_dist<float> uf;
	        while (1)
	        {
	            U = uf(streamsBeta[((*numRand)*2)]);
	            V = uf(streamsBeta[((*numRand)*2)+1]);
	            X = pow(U, 1.0/ (*aa));
	            Y = pow(V, 1.0/ (*bb));

	            if ((X + Y) <= 1.0)
	            {
	                *res = X / (X + Y);
	                return;
	            }
	        }
	    }
	    else
	    {
	    	trng::gamma_dist<> gA(1.0,*aa), gB(1.0,*bb), gAB(1.0,(*aa) + (*bb)) ;
	    	float Gab = gAB(streamsBeta[((*numRand)*2)]);
	    	float Ga = gA(streamsBeta[((*numRand)*2)]);
	    	float Gb = gB(streamsBeta[((*numRand)*2)+1]);
	        *res = Gab/(Ga + Gb);
	        //std::cout << *res << std::endl;
	        return;
	    }
}


//FUNCTION random_beta(aa, bb) RESULT(fn_val)
//
//! Adapted from Fortran 77 code from the book:
//!     Dagpunar, J. 'Principles of random variate generation'
//!     Clarendon Press, Oxford, 1988.   ISBN 0-19-852202-9
//
//! FUNCTION GENERATES A RANDOM VARIATE IN [0,1]
//! FROM A BETA DISTRIBUTION WITH DENSITY
//! PROPORTIONAL TO BETA**(AA-1) * (1-BETA)**(BB-1).
//! USING CHENG'S LOG LOGISTIC METHOD.
//
//!     AA = SHAPE PARAMETER FROM DISTRIBUTION (0 < REAL)
//!     BB = SHAPE PARAMETER FROM DISTRIBUTION (0 < REAL)
//

extern "C" void trng_random_beta_leap_(int *numRand,float*aa,float*bb,float*res) {

	if ( *aa <= 0.0 || *bb <= 0.0 ) {
		std::cerr << "trng_random_beta_leap_ :: wrong parameters aa , bb "<< *aa <<" " << *bb << std::endl ;
		return;
	}

	float a = *aa ;
	float b = *bb ;
	float aln4 = 1.3862944;
	bool  swap = b > a;
	if (swap) {
		 float g = b ;
         b = a ;
	     a = g ;
	}
	float d = a/b ;
	float f = a+b ;
	float h,t;
	if (b > 1.0 ) {
	   h = sqrt((2.0*a*b - f)/(f - 2.0));
	   t = 1.0;
	} else {
	   h = b ;
	   t = 1.0/(1.0+ pow((a/(1.0E38*b)),b)) ;
	}
	float c = a+h ;
	trng::uniform01_dist<float> uf;

	while (true ) {
 	 float r = uf(streamsBeta[((*numRand)*2)]);
 	 float x = uf(streamsBeta[((*numRand)*2)+1]);
	 float s = r*r*x ;
	 if (r <= FLT_MIN || s <= 0.0 ) continue ;
	 if (r < t) {
	    x = log(r/(1.0 - r))/h ;
	    float y = d*exp(x);
	    float z = c*x + f*log((1.0+ d)/(1.0+ y)) - aln4 ;
	    if(s - 1.0 > z) {
	       if (s - s*z > 1.0) continue ;
	       if (log(s) > z) continue;
	    }
	    *res = y/(1.0+ y) ;
	 } else {
	    if (4.0*s > pow((1.0+ 1.0/d),f)) continue;
	    *res = 1.0 ;
	 }
	 break;
	}

	if (swap) *res = 1.0 - *res ;
}
